"""
pac.py
Compute the phase-amplitude coupling between two oscillators
"""
import numpy as np
import neurodsp
import warnings
import matplotlib.pyplot as plt
from matplotlib import cm


def compute_pac(x_pha, x_amp, Fs, f_range_lo, f_range_hi,
                N_seconds_lo=None, N_seconds_hi=None,
                filter_fn=None, filter_kwargs=None,
                hilbert_increase_N=False,
                pac_method='ozkurt',
                N_bins_tort=None, N_surr_canolty=None, verbose=True):
    """
    Calculate phase-amplitude coupling between a low-frequency
    range of x_pha and a higher frequency range in x_amp

    Parameters
    ----------
    x_pha : array-like, 1d
        The time-series from which to compute the phase component
    x_amp : array-like, 1d
        The time series from which to compute the amplitude component
    Fs : float
        Sampling rate (Hz) of the two time series
    f_range_lo : tuple, 2 elements
        The low frequency filtering range (Hz)
    f_range_hi : tuple, 2 elements
        The high frequency filtering range (Hz)
    N_seconds_lo : float
        Length of the low band-pass filter (seconds)
    N_seconds_hi : float
        Length of the high band-pass filter (seconds)
    filter_fn : function or False
        The filtering function, with api:
        `filterfn(x, Fs, pass_type, f_lo, f_hi, remove_edge_artifacts=True)
        If False, it is assumed that x_pha and x_amp are the phase time
        series and the amplitude time series, respectively. Therefore, no
        filtering or hilbert transform will be done.
    filter_kwargs : dict
        Keyword parameters to pass to `filterfn(.)`
    hilbert_increase_N : bool
        if True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of x
    pac_method : str
        Indicates the method used to correlate the phase and amplitude time series in order to
        quantify the strength of pac.
        'plv': phase-locking value method (see Penny et al., 2008, J Neuro Methods)
        'glm': general linear model method (see Penny et al., 2008, J Neuro Methods)
        'tort': modulation index method (see Tort et al., 2010, J Neurophys)
        'canolty' : modulation index method (see Canolty et al., 2006, Science)
        'ozkurt' : normalized modulation index method (see Ozkurt & Schnitzler, 2011, J Neuro Methods)
    N_bins_tort : int or None
        Number of phase bins to use in Tort's modulation index method of estimating PAC
    N_surr_canolty : int or None
        Number of surrograte runs for Canolty's modulation index method of estimating PAC
    verbose : bool
        if True, print optional warning information

    Returns
    -------
    pac : float
        phase-amplitude coupling strength
    """
    # Set default filtering parameters
    if N_seconds_lo is None:
        if verbose:
            warnings.warn('Filter order not specified. Filter length automatically set to 3 cycles of the low cutoff frequency.')
        N_cycles = 3
        N_seconds_lo = N_cycles / f_range_lo[0]
    if N_seconds_hi is None:
        if verbose:
            warnings.warn('Filter order not specified. Filter length automatically set to 3 cycles of the low cutoff frequency.')
        N_cycles = 3
        N_seconds_hi = N_cycles / f_range_hi[0]
    if filter_fn is None:
        filter_fn = neurodsp.filter
    if filter_kwargs is None:
        filter_kwargs = {}

    # Only compute phase and amplitude if filter_fn is not False
    if filter_fn is not False:
        # Compute phase time series
        filter_kwargs['N_seconds'] = N_seconds_lo
        filter_kwargs['verbose'] = verbose
        pha = neurodsp.phase_by_time(x_pha, Fs, f_range_lo,
                                     filter_fn=filter_fn, filter_kwargs=filter_kwargs,
                                     hilbert_increase_N=hilbert_increase_N)

        # Compute amp time series
        filter_kwargs['N_seconds'] = N_seconds_hi
        amp = neurodsp.amp_by_time(x_amp, Fs, f_range_hi,
                                   filter_fn=filter_fn, filter_kwargs=filter_kwargs,
                                   hilbert_increase_N=hilbert_increase_N)
    else:
        # Set phase and amplitude time series to 'x' if filter_fn set to False
        pha = x_pha
        amp = x_amp

        # Reset filter function and kwargs
        filter_fn = neurodsp.filter
        filter_kwargs = {'verbose': verbose}

    # Remove the part of both signals with edge artifacts
    # The filter should be longer for the lower-frequency phase-providing
    # signal
    first_nonan = np.where(~np.isnan(pha))[0][0]
    last_nonan = np.where(~np.isnan(pha))[0][-1] + 1
    pha_nonan = pha[first_nonan:last_nonan]
    amp_nonan = amp[first_nonan:last_nonan]

    # Compute statistic relating phase and amplitude
    if pac_method == 'plv':
        pac = _plv_pac(pha, amp, Fs, f_range_lo, N_seconds_lo,
                       filter_fn, filter_kwargs, hilbert_increase_N)
    elif pac_method == 'glm':
        pac = _glm_pac(pha, amp)
    elif pac_method == 'tort':
        pac = _tort_pac(pha, amp, N_bins_tort)
    elif pac_method == 'canolty':
        pac = _canolty_pac(pha, amp, N_surr_canolty)
    elif pac_method == 'ozkurt':
        pac = _ozkurt_pac(pha, amp)
    else:
        raise ValueError('Method specified in "pac_method" not known.')
    return pac


def _plv_pac(pha, amp, Fs, f_range_lo, N_seconds_lo,
             filter_fn, filter_kwargs, hilbert_increase_N):
    """Use the PLV method to compute phase-amplitude coupling"""

    # Compute the phase of the amplitude time series
    filter_kwargs['N_seconds'] = N_seconds_lo
    amp_pha = neurodsp.phase_by_time(amp, Fs, f_range_lo,
                                     filter_fn=filter_fn, filter_kwargs=filter_kwargs,
                                     hilbert_increase_N=hilbert_increase_N)

    # Compute phase locking
    pac = np.abs(np.mean(np.exp(1j * (pha - amp_pha))))
    return pac


def _tort_pac(pha, amp, N_bins=None):
    """Use Tort's modulation index method to compute phase-amplitude coupling"""
    # Set default bin number
    if N_bins is None:
        N_bins = 20

    # Convert the phase time series from radians to degrees
    phadeg = np.degrees(pha)

    # Calculate mean amplitude in each phase bin
    binsize = 360 / N_bins
    phase_lo = np.arange(-180, 180, binsize)
    mean_amp = np.zeros(len(phase_lo))
    for b in range(len(phase_lo)):
        phaserange = np.logical_and(phadeg >= phase_lo[b],
                                    phadeg < (phase_lo[b] + binsize))
        mean_amp[b] = np.mean(amp[phaserange])

    # Compute the probability of an amplitude unit being in a phase bin
    p_j = mean_amp / sum(mean_amp)

    # Compute the entropy and PAC
    h = -np.sum(p_j * np.log10(p_j))
    h_max = np.log10(N_bins)
    pac = (h_max - h) / h_max

    return pac


def _glm_pac(pha, amp):
    """Use the GLM method to compute phase-amplitude coupling"""
    # Prepare GLM
    y = amp
    X_pre = np.vstack((np.cos(pha), np.sin(pha)))
    X = X_pre.T

    # Apply GLM and compute residuals
    y_hat, beta_hat = _ols(y, X)
    resid = y - y_hat

    # Calculate PAC from GLM residuals
    pac = 1 - np.sum(resid ** 2) / np.sum((amp - np.mean(amp)) ** 2)
    return pac


def _ols(y, X):
    """Custom OLS (to minimize outside dependecies)"""
    dummy = np.repeat(1.0, X.shape[0])
    X = np.hstack([X, dummy[:, np.newaxis]])
    beta_hat, resid, _, _ = np.linalg.lstsq(X, y)
    y_hat = np.dot(X, beta_hat)
    return y_hat, beta_hat


def _canolty_pac(pha, amp, N_surr=None):
    """Use Canolty's modulation index method to compute phase-amplitude coupling"""
    # Set default number of surrogate runs
    if N_surr is None:
        N_surr = 100

    # Calculate modulation index
    pac_raw = np.abs(np.mean(amp * np.exp(1j * pha)))

    # Calculate surrogate MIs
    pacS = np.zeros(N_surr)
    loj = np.exp(1j * pha)
    for s in range(N_surr):
        loS = np.roll(loj, np.random.randint(len(pha)))
        pacS[s] = np.abs(np.mean(amp * loS))

    # Return z-score of observed modulation index compared to null distribution
    pac = (pac_raw - np.mean(pacS)) / np.std(pacS)
    return pac


def _ozkurt_pac(pha, amp):
    """Use Ozkurt's method to compute phase-amplitude coupling"""
    # Calculate normalized modulation index
    pac = np.abs(np.sum(amp * np.exp(1j * pha))) / \
        (np.sqrt(len(pha)) * np.sqrt(np.sum(amp**2)))
    return pac


def compute_pac_comodulogram(x_pha, x_amp, Fs,
                             f_pha_bin_edges, f_amp_bin_edges,
                             N_cycles_pha=None, N_cycles_amp=None,
                             filter_fn=None, filter_kwargs=None,
                             hilbert_increase_N=False,
                             pac_method='ozkurt',
                             N_bins_tort=None, N_surr_canolty=None, verbose=True):
    """
    Calculate phase-amplitude coupling between a low-frequency
    range of x_pha and a higher frequency range in x_amp

    Parameters
    ----------
    x_pha : array-like, 1d
        The time-series from which to compute the phase component
    x_amp : array-like, 1d
        The time series from which to compute the amplitude component
    Fs : float
        Sampling rate (Hz) of the two time series
    f_pha_bin_edges: array-like, 1d
        An array of frequency values (Hz) that define the edges of the
        frequency ranges on which to estimate phase
    f_amp_bin_edges : array-like, 1d
        An array of frequency values (Hz) that define the edges of the
        frequency ranges on which to estimate amplitude
    N_cycles_pha : float
        Length of the low band-pass filter in terms of the number of cycles
        of a sine wave with a frequency at the low-cutoff of the bandpass filter
    N_cycles_amp : float
        Length of the high band-pass filter in terms of the number of cycles
        of a sine wave with a frequency at the low-cutoff of the bandpass filter
    filter_fn : function or False
        The filtering function, with api:
        `filterfn(x, Fs, pass_type, f_lo, f_hi, remove_edge_artifacts=True)
        If False, it is assumed that x_pha and x_amp are the phase time
        series and the amplitude time series, respectively. Therefore, no
        filtering or hilbert transform will be done.
    filter_kwargs : dict
        Keyword parameters to pass to `filterfn(.)`
    hilbert_increase_N : bool
        if True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of x
    pac_method : str
        Indicates the method used to correlate the phase and amplitude time series in order to
        quantify the strength of pac.
        'plv': phase-locking value method (see Penny et al., 2008, J Neuro Methods)
        'glm': general linear model method (see Penny et al., 2008, J Neuro Methods)
        'tort': modulation index method (see Tort et al., 2010, J Neurophys)
        'canolty' : modulation index method (see Canolty et al., 2006, Science)
        'ozkurt' : normalized modulation index method (see Ozkurt & Schnitzler, 2011, J Neuro Methods)
    N_bins_tort : int or None
        Number of phase bins to use in Tort's modulation index method of estimating PAC
    N_surr_canolty : int or None
        Number of surrograte runs for Canolty's modulation index method of estimating PAC
    verbose : bool
        if True, print optional warning information

    Returns
    -------
    pac : 2d array
        phase-amplitude coupling strength values for each combination of phase-providing
        frequency bin and amplitude-providing frequency bin.
    """

    # Display warning about the true width of frequency bins
    if verbose:
        warnings.warn("The true bandwidth of the filters used for each frequency bin of the comodulogram "
                      "is almost always are wider than the declared width of the frequency bin. "
                      "And this width increases as a function of frequency."
                      "For example the frequency bin 60-64Hz likely uses a bandwidth >4Hz. "
                      "You can decrease this bandwidth by increasing the N_cycles_pha and N_cycles_amp arguments. "
                      "This warning can be turned off by setting the 'verbose' kwarg to False.")

    # Set default filtering parameters
    if filter_fn is None:
        filter_fn = neurodsp.filter
    if filter_kwargs is None:
        filter_kwargs = {'N_cycles': N_cycles_pha,
                         'verbose': False}
    else:
        filter_kwargs['N_cycles'] = N_cycles_pha
        filter_kwargs['verbose'] = False

    # Compute phase time series for each frequency bin
    N_bins_pha = len(f_pha_bin_edges) - 1
    pha_by_bin = np.zeros((N_bins_pha, len(x_pha)))
    for i in range(N_bins_pha):
        f_range_temp = (f_pha_bin_edges[i], f_pha_bin_edges[i + 1])
        pha_by_bin[i] = neurodsp.phase_by_time(x_pha, Fs, f_range_temp,
                                               filter_fn=filter_fn,
                                               filter_kwargs=filter_kwargs,
                                               hilbert_increase_N=False)

    # Compute amplitude time series for each frequency bin
    N_bins_amp = len(f_amp_bin_edges) - 1
    amp_by_bin = np.zeros((N_bins_amp, len(x_pha)))
    for i in range(N_bins_amp):
        f_range_temp = (f_amp_bin_edges[i], f_amp_bin_edges[i + 1])
        amp_by_bin[i] = neurodsp.amp_by_time(x_amp, Fs, f_range_temp,
                                             filter_fn=filter_fn,
                                             filter_kwargs=filter_kwargs,
                                             hilbert_increase_N=False)

    # For each pair of frequency bins, compute PAC
    pac = np.zeros((N_bins_pha, N_bins_amp))
    for i in range(N_bins_pha):
        for j in range(N_bins_amp):
            f_range_pha_temp = (f_pha_bin_edges[i], f_pha_bin_edges[i + 1])
            f_range_amp_temp = (f_amp_bin_edges[j], f_amp_bin_edges[j + 1])
            pac[i, j] = compute_pac(pha_by_bin[i], amp_by_bin[j], Fs,
                                    f_range_pha_temp, f_range_amp_temp,
                                    filter_fn=False, pac_method=pac_method,
                                    N_bins_tort=N_bins_tort, N_surr_canolty=N_surr_canolty,
                                    verbose=False)
    return pac


def plot_pac_comodulogram(pac, f_pha_bin_edges, f_amp_bin_edges,
                          clim=None, figsize=None, colormap=None):
    """
    Plot the PAC comodulogram computed in the `compute_pac_comodulogram` function

    Parameters
    ----------
    pac : 2d array
        phase-amplitude coupling strength values for each combination of phase-providing
        frequency bin and amplitude-providing frequency bin.
    f_pha_bin_edges: array-like, 1d
        An array of frequency values (Hz) that define the edges of the
        frequency ranges on which to estimate phase
    f_amp_bin_edges : array-like, 1d
        An array of frequency values (Hz) that define the edges of the
        frequency ranges on which to estimate amplitude
    clim : 2-element tuple
        Limits in the colorbar that represents PAC strength
    figsize: 2-element tuple
        size of figure, as in the 'figsize' kwarg used in plt.figure()
    colormap: matplotlib-compatible colormap
        a colormap from matplotlib's colormap module (matplotlib.cm)
    """

    # Set defaults
    if figsize is None:
        figsize = (6, 5)
    if colormap is None:
        colormap = cm.viridis

    # Plot comodulogram
    plt.figure(figsize=figsize)
    # Set up colorbar
    if clim is None:
        cax = plt.pcolor(f_pha_bin_edges, f_amp_bin_edges, pac.T,
                         cmap=colormap)
        cbar = plt.colorbar(cax)
    else:
        cax = plt.pcolor(f_pha_bin_edges, f_amp_bin_edges, pac.T,
                         cmap=colormap, vmin=clim[0], vmax=clim[1])
        cbar = plt.colorbar(cax, ticks=clim)
        cbar.ax.set_yticklabels(clim, size=20)
    cbar.ax.set_ylabel('Modulation Index', size=20)

    # Plot labels
    plt.axis([f_pha_bin_edges[0], f_pha_bin_edges[-1],
              f_amp_bin_edges[0], f_amp_bin_edges[-1]])
    plt.xlabel('Phase frequency (Hz)', size=20)
    plt.ylabel('Amplitude frequency (Hz)', size=20)
    plt.yticks(size=20)
    plt.xticks(size=20)
    plt.tight_layout()



"""
cycles.py
Quantify the shape of oscillatory waveforms on a cycle-by-cycle basis
"""

import numpy as np
import pandas as pd
import neurodsp
import warnings


def features_by_cycle(x, Fs, f_range, center_extrema='P',
                      find_extrema_kwargs=None,
                      estimate_oscillating_periods=False,
                      true_oscillating_periods_kwargs=None,
                      hilbert_increase_N=False):
    """
    Calculate several features of an oscillation's waveform
    shape for each cycle in a recording.

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Fs : float
        sampling rate (Hz)
    f_range : (low, high), Hz
        frequency range for narrowband signal of interest
    center_extrema : str
        The center extrema in the cycle
        'P' : cycles are defined trough-to-trough
        'T' : cycles are defined peak-to-peak
    find_extrema_kwargs : dict or None
        Keyword arguments for function to find peaks and
        troughs (find_extrema) to change filter
        parameters or boundary
    estimate_oscillating_periods: bool
        if True, call _define_true_oscillating_periods to
        declare each cycle as in an oscillating or not.
    true_oscillating_periods_kwargs : dict or None
        Keyword arguments for function to find label cycles
        as in or not in an oscillation
    hilbert_increase_N : bool
        corresponding kwarg for neurodsp.amp_by_time
        If true, this zeropads the signal when computing the
        Fourier transform, which can be necessary for
        computing it in a reasonable amount of time.

    Returns
    -------
    df : pandas DataFrame
        dataframe containing several features and identifiers
        for each oscillatory cycle. Each row is one cycle.
        Each column is described below for peak-centered cycles,
        but are similar for trough-centered cycles:
        sample_peak : sample of 'x' at which the peak occurs
        sample_zerox_decay : sample of the decaying zerocrossing
        sample_zerox_rise : sample of the rising zerocrossing
        sample_last_trough : sample of the last trough
        sample_next_trough : sample of the next trough
        period : period of the cycle
        time_decay : time between peak and next trough
        time_rise : time between peak and previous trough
        time_peak : time between rise and decay zerocrosses
        time_trough : duration of previous trough estimated by zerocrossings
        volt_decay : voltage change between peak and next trough
        volt_rise : voltage change between peak and previous trough
        volt_peak : voltage at the peak
        volt_trough : voltage at the last trough
        volt_rdsym : voltage difference between rise and decay
        time_rdsym : fraction of cycle in the rise period
        volt_ptsym : voltage difference between peak and trough
        time_ptsym : fraction of cycle in the peak period
        oscillator_amplitude : average amplitude of the oscillation in that frequency band during the cycle

    Notes
    -----
    * By default, the first extrema analyzed will be a peak,
    and the final one a trough. In order to switch the preference,
    the signal is simply inverted and columns are renamed.
    """

    # Set defaults if user input is None
    if true_oscillating_periods_kwargs is None:
        true_oscillating_periods_kwargs = {}
    if find_extrema_kwargs is None:
        find_extrema_kwargs = {}
    else:
        # Raise warning if switch from peak start to trough start
        if 'first_extrema' in find_extrema_kwargs.keys():
            raise ValueError('This function has been designed to assume that\
                              the first extrema identified will be a peak.\
                              This cannot be overwritten.')

    # Negate signal if to analyze trough-centered cycles
    if center_extrema == 'P':
        pass
    elif center_extrema == 'T':
        x = -x
    else:
        raise ValueError('Parameter "center_extrema" must be either "P" or "T"')

    # Find peak and trough locations in the signal
    Ps, Ts = neurodsp.shape.find_extrema(x, Fs, f_range, **find_extrema_kwargs)

    # Find zero-crossings
    zeroxR, zeroxD = neurodsp.shape.find_zerox(x, Ps, Ts)

    # Determine number of cycles
    N_t2t = len(Ts) - 1

    # For each cycle, identify the sample of each extrema and zerocrossing
    shape_features = {}
    shape_features['sample_peak'] = Ps[1:]
    shape_features['sample_zerox_decay'] = zeroxD[1:]
    shape_features['sample_zerox_rise'] = zeroxR
    shape_features['sample_last_trough'] = Ts[:-1]
    shape_features['sample_next_trough'] = Ts[1:]

    # Compute duration of period
    shape_features['period'] = shape_features['sample_next_trough'] - \
        shape_features['sample_last_trough']

    # Compute duration of peak
    half_decay_time = (zeroxD[1:] - Ps[1:])
    half_rise_time = (Ps[1:] - zeroxR)
    shape_features['time_peak'] = half_decay_time + half_rise_time

    # Compute duration of last trough
    half_decay_time = (Ts[:-1] - zeroxD[:-1])
    half_rise_time = (zeroxR - Ts[:-1])
    shape_features['time_trough'] = half_decay_time + half_rise_time

    # Determine extrema voltage
    shape_features['volt_peak'] = x[Ps[1:]]
    shape_features['volt_trough'] = x[Ts[:-1]]

    # Determine rise and decay characteristics
    shape_features['time_decay'] = (Ts[1:] - Ps[1:])
    shape_features['time_rise'] = (Ps[1:] - Ts[:-1])

    shape_features['volt_decay'] = x[Ps[1:]] - x[Ts[1:]]
    shape_features['volt_rise'] = x[Ps[1:]] - x[Ts[:-1]]

    # Comptue rise-decay symmetry features
    shape_features['volt_rdsym'] = shape_features['volt_rise'] - shape_features['volt_decay']
    shape_features['time_rdsym'] = shape_features['time_rise'] / shape_features['period']

    # Compute peak-trough symmetry features
    shape_features['volt_ptsym'] = shape_features['volt_peak'] - shape_features['volt_trough']
    shape_features['time_ptsym'] = shape_features['time_peak'] / (shape_features['time_peak'] + shape_features['time_trough'])

    # Compute average oscillatory amplitude estimate during cycle
    amp = neurodsp.amp_by_time(x, Fs, f_range, hilbert_increase_N=hilbert_increase_N)
    shape_features['oscillator_amplitude'] = [np.mean(amp[Ts[i]:Ts[i + 1]]) for i in range(N_t2t)]

    # Convert feature dictionary into a DataFrame
    df = pd.DataFrame.from_dict(shape_features)

    # Define whether or not each cycle is part of an oscillation
    if estimate_oscillating_periods:
        df = _define_true_oscillating_periods(df, **true_oscillating_periods_kwargs)

    # Rename columns if they are actually trough-centered
    if center_extrema == 'T':
        rename_dict = {'sample_peak': 'sample_trough',
                       'sample_zerox_decay': 'sample_zerox_rise',
                       'sample_zerox_rise': 'sample_zerox_decay',
                       'sample_last_trough': 'sample_last_peak',
                       'sample_next_trough': 'sample_next_peak',
                       'time_peak': 'time_trough',
                       'time_trough': 'time_peak',
                       'volt_peak': 'volt_trough',
                       'volt_trough': 'volt_peak',
                       'time_rise': 'time_decay',
                       'time_decay': 'time_rise',
                       'volt_rise': 'volt_decay',
                       'volt_decay': 'volt_rise'}
        df.rename(columns=rename_dict, inplace=True)

        # Need to reverse symmetry measures
        df['volt_peak'] = -df['volt_peak']
        df['volt_trough'] = -df['volt_trough']
        df['volt_rdsym'] = -df['volt_rdsym']
        df['volt_ptsym'] = -df['volt_ptsym']
        df['time_rdsym'] = 1 - df['time_rdsym']
        df['time_ptsym'] = 1 - df['time_ptsym']

    return df


def _define_true_oscillating_periods(df, restrict_by_amplitude=True,
                                     restrict_by_amplitude_consistency=True,
                                     restrict_by_period_consistency=True,
                                     amplitude_fraction_threshold=.5,
                                     amplitude_consistency_threshold=.5,
                                     period_consistency_threshold=.5):
    """
    Denote which cycles in df meet the criteria to be in an oscillatory mode

    Parameters
    ----------
    df : pandas DataFrame
        dataframe of waveform features for individual cyclces
    restrict_by_amplitude : bool
        if True, require a cycle to have an amplitude above the threshold
        set in 'amplitude_fraction_threshold' in order to be considered
        in an oscillatory mode
    restrict_by_amplitude_consistency : bool
        if True, require the rise and decays within the cycle of interest
        and the two adjacent cycles to have consistent magnitudes,
        as defined by the 'amplitude_consistency_threshold' parameter
    restrict_by_period_consistency : bool
        if True, require an extrema to be adjacent to 2 cycles with consistent
        periods in order to be considered to be in an oscillatory mode
    amplitude_fraction_threshold : float (0 to 1)
        the minimum normalized amplitude a cycle must have
        in order to be considered in an oscillation.
        0 = the minimum amplitude across all cycles
        .5 = the median amplitude across all cycles
        1 = the maximum amplitude across all cycles
    amplitude_consistency_threshold : float (0 to 1)
        the minimum normalized difference in rise and decay magnitude
        to be considered as in an oscillatory mode
        1 = the same amplitude for the rise and decay
        .5 = the rise (or decay) is half the amplitude of the decay (rise)
    period_consistency_threshold : float (0 to 1)
        the minimum normalized difference in period between two adjacent cycles
        to be considered as in an oscillatory mode
        1 corresponds to the same period for both cycles
        .5 corresponds to one cycle being half the duration of another cycle

    Returns
    -------
    df : pandas DataFrame
        same df as input, except now with an extra boolean column,
        oscillating: True if that cycle met the criteria for an oscillatory mode

    Notes
    -----
    * The first and last period cannot be considered oscillating
    if the consistency measures are used.
    """

    # Make a binary array to indicate if a peak is in an oscillatory mode ("good")
    P = len(df)
    cycle_good_amp = np.zeros(P, dtype=bool)
    cycle_good_amp_consist = np.zeros(P, dtype=bool)
    cycle_good_period_consist = np.zeros(P, dtype=bool)

    # Compute if each cycle meets the amplitude reqt
    if restrict_by_amplitude:
        amps = df['oscillator_amplitude'].values
        ampnorm = (amps - np.min(amps)) / (np.max(amps) - np.min(amps))
        cycle_good_amp = ampnorm > amplitude_fraction_threshold
    else:
        cycle_good_amp = np.ones(P, dtype=bool)

    # Compute if each cycle meets the amplitude consistency reqt
    if restrict_by_amplitude_consistency:
        rises = df['volt_rise'].values
        decays = df['volt_decay'].values
        for p in range(1, P - 1):
            if cycle_good_amp[p]:
                frac1 = np.min([rises[p], decays[p]]) / np.max([rises[p], decays[p]])
                frac2 = np.min([rises[p], decays[p - 1]]) / \
                    np.max([rises[p], decays[p - 1]])
                frac3 = np.min([rises[p + 1], decays[p]]) / \
                    np.max([rises[p + 1], decays[p]])
                if np.min([frac1, frac2, frac3]) >= amplitude_consistency_threshold:
                    cycle_good_amp_consist[p] = True
    else:
        cycle_good_amp_consist = np.ones(P, dtype=bool)

    # Compute if each cycle meets the period consistency reqt
    if restrict_by_period_consistency:
        P_times = df['sample_peak'].values
        for p in range(1, P - 1):
            if np.logical_and(cycle_good_amp[p], cycle_good_amp_consist[p]):
                p1 = P_times[p] - P_times[p - 1]
                p2 = P_times[p + 1] - P_times[p]
                frac = np.min([p1, p2]) / np.max([p1, p2])
                if frac > period_consistency_threshold:
                    cycle_good_period_consist[p] = True
    else:
        cycle_good_period_consist = np.ones(P, dtype=bool)

    # Add indication for oscillating cycles to dataframe
    cycle_good = np.logical_and(cycle_good_amp, cycle_good_amp_consist)
    cycle_good = np.logical_and(cycle_good, cycle_good_period_consist)

    # Make first and last cycle nan if consistency measure used
    if restrict_by_amplitude_consistency or restrict_by_period_consistency:
        cycle_good[0] = np.nan
        cycle_good[-1] = np.nan

    # Add oscillating detector to the dataframe
    df['oscillating'] = cycle_good
    return df



"""
cyclespoints.py
Functions to determine the locations of peaks, troughs, and zerocrossings (rise and decay)
for individual cycles
"""

import numpy as np
import neurodsp
import warnings


def find_extrema(x, Fs, f_range, boundary=None, first_extrema='peak',
                 filter_fn=None, filter_kwargs=None):
    """
    Identify peaks and troughs in a time series.

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Fs : float
        sampling rate
    f_range : (low, high), Hz
        frequency range for narrowband signal of interest,
        used to find zerocrossings of the oscillation
    boundary : int
        number of samples from edge of recording to ignore
    first_extrema: str or None
        if 'peak', then force the output to begin with a peak and end in a trough
        if 'trough', then force the output to begin with a trough and end in peak
        if None, force nothing
    filter_fn : filter function, `filterfn(x, Fs, pass_type, f_lo, f_hi, remove_edge_artifacts=True)
        Must have the same API as neurodsp.filter
    filter_kwargs : dict
        keyword arguments to the filter_fn

    Returns
    -------
    Ps : array-like 1d
        indices at which oscillatory peaks occur in the input signal x
    Ts : array-like 1d
        indices at which oscillatory troughs occur in the input signal x

    Notes
    -----
    This function assures that there are the same number of peaks and troughs
    if the first extrema is forced to be either peak or trough.
    """

    # Set default filtering parameters
    if filter_fn is None:
        filter_fn = neurodsp.filter
    if filter_kwargs is None:
        filter_kwargs = {}

    # Default boundary value as 1 cycle length
    if boundary is None:
        boundary = int(np.ceil(Fs / float(f_range[0])))

    # Filter signal
    x_filt = filter_fn(x, Fs, 'bandpass', f_lo=f_range[0], f_hi=f_range[1], remove_edge_artifacts=False, **filter_kwargs)

    # Find rising and falling zerocrossings
    zeroriseN = _fzerorise(x_filt)
    zerofallN = _fzerofall(x_filt)

    # Compute number of peaks and troughs
    if zeroriseN[-1] > zerofallN[-1]:
        P = len(zeroriseN) - 1
        T = len(zerofallN)
    else:
        P = len(zeroriseN)
        T = len(zerofallN) - 1

    # Calculate peak samples
    Ps = np.zeros(P, dtype=int)
    for p in range(P):
        # Calculate the sample range between the most recent zero rise
        # and the next zero fall
        mrzerorise = zeroriseN[p]
        nfzerofall = zerofallN[zerofallN > mrzerorise][0]
        # Identify time fo peak
        Ps[p] = np.argmax(x[mrzerorise:nfzerofall]) + mrzerorise

    # Calculate trough samples
    Ts = np.zeros(T, dtype=int)
    for tr in range(T):
        # Calculate the sample range between the most recent zero fall
        # and the next zero rise
        mrzerofall = zerofallN[tr]
        nfzerorise = zeroriseN[zeroriseN > mrzerofall][0]
        # Identify time of trough
        Ts[tr] = np.argmin(x[mrzerofall:nfzerorise]) + mrzerofall

    # Remove peaks and troughs within the boundary limit
    Ps = Ps[np.logical_and(Ps > boundary, Ps < len(x) - boundary)]
    Ts = Ts[np.logical_and(Ts > boundary, Ts < len(x) - boundary)]

    # Force the first extrema to be as desired
    # Assure equal # of peaks and troughs
    if first_extrema == 'peak':
        if Ps[0] > Ts[0]:
            Ts = Ts[1:]
        if Ps[-1] > Ts[-1]:
            Ps = Ps[:-1]
    elif first_extrema == 'trough':
        if Ts[0] > Ps[0]:
            Ps = Ps[1:]
        if Ts[-1] > Ps[-1]:
            Ts = Ts[:-1]
    elif first_extrema is None:
        pass
    else:
        raise ValueError('Parameter forcestart is invalid')

    return Ps, Ts


def _fzerofall(data):
    """Find zerocrossings on falling edge of a filtered signal"""
    pos = data > 0
    return (pos[:-1] & ~pos[1:]).nonzero()[0]


def _fzerorise(data):
    """Find zerocrossings on rising edge of a filtered signal"""
    pos = data < 0
    return (pos[:-1] & ~pos[1:]).nonzero()[0]


def find_zerox(x, Ps, Ts):
    """
    Find zerocrossings within each cycle after peaks and troughs are identified.
    A rising zerocrossing occurs when the voltage crosses
    midway between the trough voltage and subsequent peak voltage.
    A decay zerocrossing is defined similarly.
    If this voltage is crossed at multiple times, the temporal median is taken
    as the zerocrossing.

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Ps : numpy arrays 1d
        time points of oscillatory peaks
    Ts : numpy arrays 1d
        time points of osillatory troughs

    Returns
    -------
    zeroxR : array-like 1d
        indices at which oscillatory rising zerocrossings occur
    zeroxD : array-like 1d
        indices at which oscillatory decaying zerocrossings occur

    Notes
    -----
    * Sometimes, due to noise in estimating peaks and troughs when the oscillation
    is absent, the estimated peak might be lower than an adjacent trough. If this
    occurs, the rise and decay zerocrossings will be set to be halfway between
    the peak and trough.
    """

    # Calculate the number of rises and decays
    if Ps[0] < Ts[0]:
        N_rises = len(Ps) - 1
        N_decays = len(Ts)
        idx_bias = 0
    else:
        N_rises = len(Ps)
        N_decays = len(Ts) - 1
        idx_bias = 1

    # Find zerocrossings for rise
    zeroxR = np.zeros(N_rises, dtype=int)
    for i in range(N_rises):
        x_temp = np.copy(x[Ts[i]:Ps[i + 1 - idx_bias] + 1])
        x_temp -= (x_temp[0] + x_temp[-1]) / 2.

        # Catch if rise is actually a net decay
        if x_temp[0] > x_temp[-1]:
            zeroxR[i] = Ts[i] + int(len(x_temp) / 2.)
        else:
            try:
                zeroxR[i] = Ts[i] + int(np.median(_fzerorise(x_temp)))
            except:
                warnings.warn('Error when estimating rising zerocrossing after trough ' + str(i) +
                              ' at sample ' + str(Ts[i]) +
                              '. Therefore, the zerocrossing has been set to halfway between the two extrema.' +
                              ' This is potentially due to the two extrema being too close together')
                zeroxR[i] = Ts[i] + int(len(x_temp) / 2.)

    # Find zerocrossings for decays
    zeroxD = np.zeros(N_decays, dtype=int)
    for i in range(N_decays):
        x_temp = np.copy(x[Ps[i]:Ts[i + idx_bias] + 1])
        x_temp -= (x_temp[0] + x_temp[-1]) / 2.

        # Catch if the decay period is actually a net rise
        if x_temp[0] < x_temp[-1]:
            zeroxD[i] = Ps[i] + int(len(x_temp) / 2.)
        else:
            try:
                zeroxD[i] = Ps[i] + int(np.median(_fzerofall(x_temp)))
            except:
                warnings.warn('Error when estimating decaying zerocrossing after peak ' + str(i) +
                              ' at sample ' + str(Ps[i]) +
                              '. Therefore, the zerocrossing has been set to halfway between the two extrema.')
                zeroxD[i] = Ps[i] + int(len(x_temp) / 2.)

    return zeroxR, zeroxD



"""
spectral.py
Frequency domain analysis of neural signals: creating PSD, fitting 1/f, spectral histograms
"""

import numpy as np
from scipy import signal
import matplotlib.pylab as plt
from sklearn import linear_model


def psd(x, Fs, method='mean', window='hann', nperseg=None, noverlap=None, filtlen=1.):
    """
    Estimating the power spectral density (PSD) of a time series from short-time Fourier
    Transform (mean, median), or the entire signal's FFT smoothed (medfilt)

    Parameters
    -----------
    x : array_like 1d
            Time series of measurement values
    Fs : float, Hz
            Sampling frequency of the x time series.
    method : { 'mean', 'median', 'medfilt'}, optional
            Methods to calculate the PSD. 'mean' is the same as Welch's method (mean of STFT), 'median' uses median of STFT instead of mean to minimize outlier effect. 'medfilt' filters the entire signals raw FFT with a median filter to smooth. Defaults to 'mean'.
    The next 3 parameters are only relevant for method = {'mean', 'median'}
    window : str or tuple or array_like, optional
            Desired window to use. See scipy.signal.get_window for a list of windows and required parameters. If window is array_like it will be used directly as the window and its length must be nperseg. Defaults to a Hann window.
    nperseg : int, optional
            Length of each segment. Defaults to None, but if window is str or tuple, is set to 1 second of data, and if window is array_like, is set to the length of the window.
    noverlap : int, optional
            Number of points to overlap between segments. If None, noverlap = nperseg // 2. Defaults to None.
    filten : float, Hz, optional
                        (For medfilt method) Length of median filter in Hz.

    Returns
    -------
    freq : ndarray
            Array of sample frequencies.
    Pxx : ndarray
            Power spectral density of x.

    References
    ----------
    Mostly relies on scipy.signal.spectrogram and numpy.fft
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html
    """

    if method in ('mean', 'median'):
        # welch-style spectrum (mean/median of STFT)
        if nperseg is None:
            if type(window) in (str, tuple):
                # window is a string or tuple, defaults to 1 second of data
                nperseg = int(Fs)
            else:
                # window is an array, defaults to window length
                nperseg = len(window)
        else:
            nperseg = int(nperseg)

        if noverlap is not None:
            noverlap = int(noverlap)

        # call signal.spectrogram function in scipy to compute STFT
        freq, t, spg = signal.spectrogram(x, Fs, window, nperseg, noverlap)
        if method is 'mean':
            Pxx = np.mean(spg, axis=-1)
        elif method is 'median':
            Pxx = np.median(spg, axis=-1)

    elif method is 'medfilt':
        # median filtered FFT spectrum
        # take the positive half of the spectrum since it's symmetrical
        FT = np.fft.fft(x)[:int(np.ceil(len(x) / 2.))]
        freq = np.fft.fftfreq(len(x), 1. / Fs)[:int(np.ceil(len(x) / 2.))]  # get freq axis
        # convert median filter length from Hz to samples
        filtlen_samp = int(int(filtlen / (freq[1] - freq[0])) / 2 * 2 + 1)
        Pxx = signal.medfilt(np.abs(FT)**2. / (Fs * len(x)), filtlen_samp)

    else:
        raise ValueError('Unknown PSD method: %s' % method)

    return freq, Pxx


def scv(x, Fs, window='hann', nperseg=None, noverlap=0, outlierpct=None):
    """
    Compute the spectral coefficient of variation (SCV) at each frequency.
    White noise should have a SCV of 1 at all frequencies.

    Parameters
    -----------
    x : array_like 1d
            Time series of measurement values
    Fs : float, Hz
            Sampling frequency of the x time series.
    window : str or tuple or array_like, optional
            Desired window to use. See scipy.signal.get_window for a list of windows and required parameters. If window is array_like it will be used directly as the window and its length must be nperseg. Defaults to a Hann window.
    nperseg : int, optional
            Length of each segment. Defaults to None, but if window is str or tuple, is set to 1 second of data, and if window is array_like, is set to the length of the window.
    noverlap : int, optional
            Number of points to overlap between segments. Defaults to 0 for independence.
    outlierpct : float, percent, optional
            Discarding a percentage of the windows with the lowest and highest total log power.

    Returns
    -------
    freq : ndarray
            Array of sample frequencies.
    SCV : ndarray
            Spectral coefficient of variation.
    """
    if nperseg is None:
        if type(window) in (str, tuple):
            # window is a string or tuple, defaults to 1 second of data
            nperseg = int(Fs)
        else:
            # window is an array, defaults to window length
            nperseg = len(window)
    else:
        nperseg = int(nperseg)

    if noverlap is not None:
        noverlap = int(noverlap)

    freq, t, spg = signal.spectrogram(x, Fs, window, nperseg, noverlap)
    if outlierpct is not None:
        # discard time windows with very low and very high powers
        discard = int(spg.shape[1] / 100. * outlierpct)
        outlieridx = np.argsort(np.mean(np.log10(spg), axis=0))[discard:-discard]
        spg = spg[:, outlieridx]

    spectcv = np.std(spg, axis=-1) / np.mean(spg, axis=-1)
    return freq, spectcv


def scv_rs(x, Fs, window='hann', nperseg=None, noverlap=0, method='bootstrap', rs_params=None):
    """
    Resampled version of scv: instead of a single estimate of mean and standard deviation, the spectrogram is resampled, either randomly (bootstrap) or time-stepped (rolling).

    Parameters
    -----------
    x : array_like 1d
            Time series of measurement values
    Fs : float, Hz
            Sampling frequency of the x time series.
    window : str or tuple or array_like, optional
            Desired window to use. See scipy.signal.get_window for a list of windows and required parameters. If window is array_like it will be used directly as the window and its length must be nperseg. Defaults to a Hann window.
    nperseg : int, optional
            Length of each segment. Defaults to None, but if window is str or tuple, is set to 1 second of data, and if window is array_like, is set to the length of the window.
    noverlap : int, optional
            Number of points to overlap between segments. Defaults to 0 for independence.
    method : {'bootstrap', 'rolling'}, optional
            Method of resampling: bootstrap randomly samples a subset of the spectrogram repeatedly, while rolling takes the rolling window scv.
            Defaults to 'bootstrap'.
    rs_params : tuple, (int, int), optional
            Parameters for resampling algorithm.
            If 'bootstrap', rs_params = (nslices, ndraws), representing number of slices per draw, and number of random draws, defaults to (10% of slices, 100 draws).
            If 'rolling', rs_params = (nslices, nsteps), representing number of slices per draw, and number of slices to step forward, defaults to (10, 5).
    Returns
    -------
    freq : ndarray
            Array of sample frequencies.
    T : ndarray
            Array of time indices, for 'rolling' resampling. If 'bootstrap', T = None.
    spectcv_rs : ndarray
            Resampled spectral coefficient of variation.
    """
    if nperseg is None:
        if type(window) in (str, tuple):
            # window is a string or tuple, defaults to 1 second of data
            nperseg = int(Fs)
        else:
            # window is an array, defaults to window length
            nperseg = len(window)
    else:
        nperseg = int(nperseg)

    if noverlap is not None:
        noverlap = int(noverlap)
    # compute spectrogram
    freq, t, spg = signal.spectrogram(x, Fs, window, nperseg, noverlap)

    if method is 'bootstrap':
        # params are number of slices of STFT to compute SCV over, and number of draws
        if rs_params is None:
            # defaults to draw 1/10 of STFT slices, 100 draws
            rs_params = (int(spg.shape[1] / 10.), 100)

        nslices, ndraws = rs_params
        spectcv_rs = np.zeros((len(freq), ndraws))
        for draw in range(ndraws):
            # repeated subsampling of spectrogram randomly, with replacement between draws
            idx = np.random.choice(spg.shape[1], size=nslices, replace=False)
            spectcv_rs[:, draw] = np.std(spg[:, idx], axis=-1) / np.mean(spg[:, idx], axis=-1)

        T = None  # no time component, return nothing

    elif method is 'rolling':
        # params are number of slices of STFT to compute SCV over, and number of slices to roll forward
        if rs_params is None:
            # defaults to 10 STFT slices, move forward by 5 slices
            rs_params = (10, 5)

        nslices, nsteps = rs_params
        outlen = int(np.ceil((spg.shape[1] - nslices) / float(nsteps))) + 1
        spectcv_rs = np.zeros((len(freq), outlen))
        for ind in range(outlen):
            curblock = spg[:, nsteps * ind:nslices + nsteps * ind]
            spectcv_rs[:, ind] = np.std(curblock, axis=-1) / np.mean(curblock, axis=-1)

        T = t[0::nsteps]  # grab the time indices from the spectrogram

    else:
        raise ValueError('Unknown resampling method: %s' % method)

    return freq, T, spectcv_rs


def spectral_hist(x, Fs, window='hann', nperseg=None, noverlap=None, nbins=50, flim=(0., 100.), cutpct=(0., 100.)):
    """
    Compute the distribution of log10 power at each frequency from the signal spectrogram.
        The histogram bins are the same for every frequency, thus evenly spacing the global min and max power

    Parameters
    -----------
    x : array_like 1d
            Time series of measurement values
    Fs : float, Hz
            Sampling frequency of the x time series.
    window : str or tuple or array_like, optional
            Desired window to use. See scipy.signal.get_window for a list of windows and required parameters. If window is array_like it will be used directly as the window and its length must be nperseg. Defaults to a Hann window.
    nperseg : int, optional
            Length of each segment. Defaults to None, but if window is str or tuple, is set to 1 second of data, and if window is array_like, is set to the length of the window.
    noverlap : int, optional
            Number of points to overlap between segments. If None, noverlap = nperseg // 2. Defaults to None.
    nbins : int, optional
            Number of histogram bins to use, defaults to 50
    flim : tuple, (start, end) in Hz, optional
            Frequency range of the spectrogram across which to compute the histograms. Default to (0., 100.)
    cutpct : tuple, (low, high), in percentage, optional
                Power percentile at which to draw the lower and upper bin limits. Default to (0., 100.)

    Returns
    -------
    freq : ndarray
            Array of frequencies.
    power_bins : ndarray
            Histogram bins used to compute the distribution.
    spect_hist : ndarray (2D)
            Power distribution at every frequency, nbins x freqs 2D matrix
    """

    if nperseg is None:
        if type(window) in (str, tuple):
            # window is a string or tuple, defaults to 1 second of data
            nperseg = int(Fs)
        else:
            # window is an array, defaults to window length
            nperseg = len(window)
    else:
        nperseg = int(nperseg)

    if noverlap is not None:
        noverlap = int(noverlap)

    # compute spectrogram of data
    freq, t, spg = signal.spectrogram(x, Fs, window, nperseg, noverlap, return_onesided=True)

    # get log10 power before binning
    ps = np.transpose(np.log10(spg))

    # Limit spectrogram to freq range of interest
    ps = ps[:, np.logical_and(freq >= flim[0], freq < flim[1])]
    freq = freq[np.logical_and(freq >= flim[0], freq < flim[1])]

    # Prepare bins for power. Min and max of bins determined by power cutoff percentage
    power_min, power_max = np.percentile(np.ndarray.flatten(ps), cutpct)
    power_bins = np.linspace(power_min, power_max, nbins + 1)

    # Compute histogram of power for each frequency
    spect_hist = np.zeros((len(ps[0]), nbins))
    for i in range(len(ps[0])):
        spect_hist[i], _ = np.histogram(ps[:, i], power_bins)
        spect_hist[i] = spect_hist[i] / sum(spect_hist[i])

    # flip them for sensible plotting direction
    spect_hist = np.transpose(spect_hist)
    spect_hist = np.flipud(spect_hist)

    return freq, power_bins, spect_hist


def plot_spectral_hist(freq, power_bins, spect_hist, psd_freq=None, psd=None):
    """
    Plot the spectral histogram.

    Parameters
    ----------
    freq : array_like, 1d
            Frequencies over which the histogram is calculated.
    power_bins : array_like, 1d
            Power bins within which histogram is aggregated.
    spect_hist : ndarray, 2d
            Spectral histogram to be plotted.
    psd_freq : array_like, 1d, optional
            Frequency axis of the PSD to be plotted.
    psd : array_like, 1d, optional
            PSD to be plotted over the histograms.
    """
    # automatically scale figure height based on number of bins
    plt.figure(figsize=(8, 12 * len(power_bins) / len(freq)))
    # plot histogram intensity as image and automatically adjust aspect ratio
    plt.imshow(spect_hist, extent=[freq[0], freq[-1], power_bins[0], power_bins[-1]], aspect='auto')
    plt.xlabel('Frequency (Hz)', fontsize=15)
    plt.ylabel('Log10 Power', fontsize=15)
    plt.colorbar(label='Probability')

    if psd is not None:
        # if a PSD is provided, plot over the histogram data
        plt.plot(psd_freq[np.logical_and(psd_freq >= freq[0], psd_freq <= freq[-1])], np.log10(
            psd[np.logical_and(psd_freq >= freq[0], psd_freq <= freq[-1])]), color='w', alpha=0.8)


def fit_slope(freq, psd, fit_frange, fit_excl=None, method='ols', plot_fit=False):
    """
    Fit PSD with straight line in log-log domain over the specified frequency range.

        Parameters
        ----------
        freq : array_like, 1d
                        Frequency axis of PSD
        psd : array_like, 1d
                        PSD to be fit over
        fit_frange : tuple, (start, end), Hz
                        Frequency range to be fit over, in Hz, inclusive on both ends.
        fit_excl : list of tuples, [(start, end), (start, end), ...], Hz, optional
                        Frequency ranges to be excluded from fit. Each element in list describes
                        the start and end of an exclusion zone.
        method : str, {'ols', 'RANSAC'}, optional
                        Line fitting method. Defaults to 'ols'
                        'ols' is ordinary least squares fit with polyfit.
                        'RANSAC' is iterative robust fit discarding outliers.
        plot_fit : bool, optional
                        If True, the PSD is plotted, along with the actual fitted PSD (excluding exclusion freqs),
                        as well as the fitted line itself. Defaults to False.

        Returns
        -------
        slope : float
                Slope of loglog fitted line, m in y = mx+b
        offset : float
                Offset of loglog fitted line, b in y = mx+b
    """

    # make a mask for included and excluded frequency regions
    fmask = np.zeros_like(freq)
    # get freq indices within the fit frequencies
    fmask[np.logical_and(freq >= fit_frange[0], freq <= fit_frange[1])] = 1
    # discard freq indices within the exclusion frequencies
    if fit_excl is not None:
        # if a tuple is given, convert it to a list
        if type(fit_excl) is tuple:
            fit_excl = [fit_excl]

        for exc_frange in fit_excl:
            fmask[np.logical_and(freq >= exc_frange[0], freq <= exc_frange[1])] = 0

    # grab the psd and freqs to be fit over
    logf = np.log10(freq[fmask == 1])
    logpsd = np.log10(psd[fmask == 1])

    # fit line
    if method is 'ols':
        # solve regular least square
        slope, offset = np.polyfit(logf, logpsd, deg=1)

    elif method is 'RANSAC':
        lm = linear_model.RANSACRegressor(random_state=42)
        lm.fit(logf.reshape(-1, 1), logpsd.reshape(-1, 1))
        offset = lm.predict(0.)[0][0]
        slope = lm.estimator_.coef_[0][0]

    else:
        raise ValueError('Unknown PSD fitting method: %s' % method)

    if plot_fit:
        plt.figure(figsize=(5, 5))
        plt.plot(np.log10(freq), np.log10(psd), label='Whole PSD')
        plt.plot(logf, logpsd, '-o', label='Fitted PSD', alpha=0.4)
        plt.plot(logf, logf * slope + offset, '-k', label='Fit Line', lw=3)
        plt.legend()
        plt.xlabel('Log10 Frequency (Hz)', fontsize=15)
        plt.ylabel('Log10 Power (V^2/Hz)', fontsize=15)

    return slope, offset



"""
test_pac.py
Test functions in the phase-amplitude coupling module
"""

import numpy as np
import os
import neurodsp
from neurodsp.tests import _load_example_data


def test_pac_consistent():
    """
    Confirm consistency in estimation of pac
    with computations in previous versions
    """

    # Compute pacs
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000

    # Compute pacs
    all_pac_methods = ['ozkurt', 'plv', 'glm', 'tort', 'canolty']
    f_range_lo = (13, 30)
    f_range_hi = (50, 200)
    N_seconds_lo = .25
    N_seconds_hi = .2

    np.random.seed(0)
    pacs = np.zeros(len(all_pac_methods))
    for i, m in enumerate(all_pac_methods):
        pacs[i] = neurodsp.compute_pac(x, x, Fs, f_range_lo, f_range_hi,
                                       N_seconds_lo=N_seconds_lo, N_seconds_hi=N_seconds_hi,
                                       pac_method=m)

    # Load ground truth pacs
    pacs_true = np.load(os.path.dirname(neurodsp.__file__) +
                        '/tests/data/sample_data_' + str(data_idx) + '_pacs.npy')

    # Compute difference between current and past signals
    assert np.allclose(np.sum(np.abs(pacs - pacs_true)), 0, atol=10 ** -5)


def test_comodulogram_consistent():
    """
    Confirm consistency in estimation of pac comodulogram
    with computations in previous versions
    """

    # Compute pacs
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000

    # Parameters for comodulogram function
    f_pha_bin_edges = np.arange(2, 42, 2)
    f_amp_bin_edges = np.arange(20, 200, 4)
    N_cycles_pha = 5
    N_cycles_amp = 11

    # Compute comodulogram
    comod = neurodsp.compute_pac_comodulogram(x, x, Fs,
                                              f_pha_bin_edges, f_amp_bin_edges,
                                              N_cycles_pha=N_cycles_pha,
                                              N_cycles_amp=N_cycles_amp,
                                              pac_method='ozkurt')

    # Load ground truth comodulogram
    comod_true = np.load(os.path.dirname(neurodsp.__file__) +
                         '/tests/data/sample_data_' + str(data_idx) + '_comod.npy')

    # Compute difference between current and past signals
    assert np.allclose(np.sum(np.abs(comod - comod_true)), 0, atol=10 ** -5)



"""
test_cyclefeatures.py
Test measurement of cycle-by-cycle features of oscillatory waveforms
"""

import numpy as np
import pandas as pd
import os
import neurodsp
from neurodsp import shape
from neurodsp.tests import _load_example_data


def test_cyclefeatures_consistent():
    """
    Confirm consistency in peak finding
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000
    f_range = (13, 30)

    # Load ground truth lagged coherence
    df_true = pd.read_csv(os.path.dirname(neurodsp.__file__) +
                          '/tests/data/sample_data_' + str(data_idx) + '_cyclefeatures.csv')

    # Compute lagged coherence
    true_oscillating_periods_kwargs = {'restrict_by_amplitude_consistency': False,
                                       'restrict_by_period_consistency': False,
                                       'amplitude_fraction_threshold': .3}

    df = shape.features_by_cycle(x, Fs, f_range, center_extrema='T',
                                 estimate_oscillating_periods=True,
                                 true_oscillating_periods_kwargs=true_oscillating_periods_kwargs)

    # Compute difference between calculated and ground truth values for each column
    for k in df.keys():
        signal_diff = df[k].values - df_true[k].values
        assert np.allclose(np.sum(np.abs(signal_diff)), 0, atol=10 ** -5)



"""
test_laggedcoherence.py
Test function to compute lagged coherence
"""

import numpy as np
import os
import neurodsp
from neurodsp.tests import _load_example_data


def test_laggedcoherence_consistent():
    """
    Confirm consistency in beta bandpass filter results on a neural signal
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000
    f_range = (13, 30)

    # Load ground truth lagged coherence
    lc_true = np.load(os.path.dirname(neurodsp.__file__) +
                      '/tests/data/sample_data_' + str(data_idx) + '_laggedcoherence.npy')

    # Compute lagged coherence
    lag_coh_beta = neurodsp.lagged_coherence(x, f_range, Fs)

    assert np.allclose(lag_coh_beta, lc_true, atol=10 ** -5)



"""
swm.py
Identify the waveform shape of neural oscillations using the sliding window matching algorithm
"""

import numpy as np


def sliding_window_matching(x, Fs, L, G,
                            max_iterations=500, T=1, window_starts_custom=None):
    """
    Find recurring patterns in a time series using the
    sliding window matching algorithm

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Fs : float
        sampling rate (samples per second)
    L : float
        window length (seconds)
    G : float
        minimum window spacing (seconds)
    T : float
        temperature parameter. Controls probability of accepting a new window
    max_iterations : int
        Maximum number of iterations of potential changes in window placement
    window_starts_custom : np.ndarray (1d)
        Pre-set locations of initial windows (instead of evenly spaced by 2G)

    Returns
    -------
    avg_window : np.ndarray (1d)
        the average waveform in 'x' in the frequency 'f_range' triggered on 'trigger'
    window_starts : np.ndarray (1d)
        indices at which each window begins for the final set of windows
    J : np.ndarray (1d)
        Cost function value at each iteration

    References
    ----------
    Gips, B., Bahramisharif, A., Lowet, E., Roberts, M. J., de Weerd, P.,
    Jensen, O., & van der Eerden, J. (2017). Discovering recurring
    patterns in electrophysiological recordings.
    Journal of Neuroscience Methods, 275, 66-79.
    MATLAB code: https://github.com/bartgips/SWM

    Notes
    -----
    * Apply a highpass filter if looking at high frequency activity,
      so that it does not converge on a low frequency motif
    * L and G should be chosen to be about the size of the motif of interest,
       and the N derived should be about the number of occurrences
    """

    # Compute window length and spacing in samples
    L_samp = int(L * Fs)
    G_samp = int(G * Fs)

    # Initialize window positions, separated by 2*G
    if window_starts_custom is None:
        window_starts = np.arange(0, len(x) - L_samp, 2 * G_samp)
    else:
        window_starts = window_starts_custom
    N_windows = len(window_starts)

    # Calculate initial cost
    J = np.zeros(max_iterations)
    J[0] = _compute_J(x, window_starts, L_samp)

    # Randomly sample windows with replacement
    random_window_idx = np.random.choice(range(N_windows), size=max_iterations)

    # For each iteration, randomly replace a window with a new window
    # to improve cross-window similarity
    iter_num = 1
    while iter_num < max_iterations:
        # Pick a random window position
        window_idx_replace = random_window_idx[iter_num]

        # Find a new allowed position for the window
        window_starts_temp = np.copy(window_starts)
        window_starts_temp[window_idx_replace] = _find_new_windowidx(
            window_starts, G_samp, L_samp, len(x) - L_samp)

        # Calculate the cost
        J_temp = _compute_J(x, window_starts_temp, L_samp)

        # Calculate the change in cost function
        deltaJ = J_temp - J[iter_num - 1]

        # Calculate the acceptance probability
        p_accept = np.exp(-deltaJ / float(T))

        # Accept update to J with a certain probability
        if np.random.rand() < p_accept:
            # Update J
            J[iter_num] = J_temp
            # Update X
            window_starts = window_starts_temp
        else:
            # Update J
            J[iter_num] = J[iter_num - 1]

        # Update iteration number
        iter_num += 1

    # Calculate average window
    avg_window = np.zeros(L_samp)
    for w in range(N_windows):
        avg_window = avg_window + x[window_starts[w]:window_starts[w] + L_samp]
    avg_window = avg_window / float(N_windows)

    return avg_window, window_starts, J


def _compute_J(x, window_starts, L_samp):
    """Compute the cost, which is proportional to the
    difference between pairs of windows"""

    # Get all windows and zscore them
    N_windows = len(window_starts)
    windows = np.zeros((N_windows, L_samp))
    for w in range(N_windows):
        temp = x[window_starts[w]:window_starts[w] + L_samp]
        windows[w] = (temp - np.mean(temp)) / np.std(temp)

    # Calculate distances for all pairs of windows
    d = []
    for i in range(N_windows):
        for j in range(i + 1, N_windows):
            window_diff = windows[i] - windows[j]
            d_temp = np.sum(window_diff**2) / float(L_samp)
            d.append(d_temp)
    # Calculate cost, the average difference, roughly
    J = np.sum(d) / float(2 * (N_windows - 1))
    return J


def _find_new_windowidx(window_starts, G_samp, L_samp, N_samp, tries_limit=1000):
    """Find a new sample for the starting window"""

    found = False
    N_tries = 0
    while found is False:
        # Generate a random sample
        new_samp = np.random.randint(N_samp)
        # Check how close the sample is to other window starts
        dists = np.abs(window_starts - new_samp)
        if np.min(dists) > G_samp:
            return new_samp
        else:
            N_tries += 1
            if N_tries > tries_limit:
                raise RuntimeError('SWM algorithm has difficulty finding a new window. Increase the spacing parameter, G.')



"""
test_filt.py
Test filtering functions
"""

import pytest
import numpy as np
import neurodsp
from neurodsp.tests import _load_example_data, _generate_random_x


def test_bandpass_filter_consistent():
    """
    Confirm consistency in beta bandpass filter results on a neural signal
    """
    # Load data and ground-truth filtered signal
    x, x_filt_true = _load_example_data(data_idx=1, filtered=True)

    # filter data
    Fs = 1000
    f_lo = 13
    f_hi = 30
    x_filt = neurodsp.filter(x, Fs, 'bandpass', f_lo=f_lo, f_hi=f_hi, N_cycles=3)

    # Compute difference between current and past filtered signals
    signal_diff = x_filt[~np.isnan(x_filt)] - x_filt_true[~np.isnan(x_filt_true)]
    assert np.allclose(np.sum(np.abs(signal_diff)), 0, atol=10 ** -5)


def test_edge_nan():
    """
    Confirm that the appropriate amount of edge artifact has been removed for FIR filters
    and that edge artifacts are not removed for IIR filters
    """

    # Apply a 4-8Hz bandpass filter to random noise
    x = _generate_random_x()
    x_filt, kernel = neurodsp.filter(x, 1000, 'bandpass', f_lo=4, f_hi=8, return_kernel=True)

    # Check if the correct edge artifacts have been removed
    N_rmv = int(np.ceil(len(kernel) / 2))
    assert all(np.isnan(x_filt[:N_rmv]))
    assert all(np.isnan(x_filt[-N_rmv:]))
    assert all(~np.isnan(x_filt[N_rmv:-N_rmv]))

    # Check that no edge artifacts are removed for IIR filters
    x_filt = neurodsp.filter(x, 1000, 'bandpass', f_lo=4, f_hi=8, iir=True, butterworth_order=3)
    assert all(~np.isnan(x_filt))


def test_frequency_input_errors():
    """
    Check that errors are properly raised when incorrectly enter frequency information
    """

    # Generate a random signal
    x = _generate_random_x()

    # Check that a bandpass filter cannot be completed without proper frequency limits
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'bandpass', f_lo=4)
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'bandpass', f_hi=8)
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'bandpass', f_lo=8, f_hi=4)

    # Check that a bandstop filter cannot be completed without proper frequency limits
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'bandstop', f_lo=58)
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'bandstop', f_hi=62)
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'bandstop', f_lo=62, f_hi=58)

    # Check that a lowpass filter cannot be completed without proper frequency limits
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'lowpass', f_hi=100)
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'lowpass', f_lo=10, f_hi=100)

    # Check that a highpass filter cannot be completed without proper frequency limits
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'highpass', f_lo=100)
    with pytest.raises(ValueError):
        x_filt = neurodsp.filter(x, 1000, 'highpass', f_lo=10, f_hi=100)


def test_filter_length():
    """
    Check that the output kernel is of the correct length
    """

    # Generate a random signal
    x = _generate_random_x()

    # Specify filter length with number of cycles
    Fs = 1000
    f_lo = 4
    f_hi = 8
    N_cycles = 5
    x_filt, kernel = neurodsp.filter(x, Fs, 'bandpass', f_lo=f_lo,
                                     f_hi=f_hi, N_cycles=N_cycles, return_kernel=True)

    # Compute how long the kernel should be
    force_kernel_length = int(np.ceil(Fs * N_cycles / f_lo))
    if force_kernel_length % 2 == 0:
        force_kernel_length = force_kernel_length + 1

    # Check correct length when defining number of cycles
    assert np.allclose(len(kernel), force_kernel_length, atol=.1)

    # Specify filter length with number of seconds
    Fs = 1000
    f_lo = 4
    f_hi = 8
    N_seconds = .8
    x_filt, kernel = neurodsp.filter(x, Fs, 'bandpass', f_lo=f_lo,
                                     f_hi=f_hi, N_seconds=N_seconds, return_kernel=True)

    # Compute how long the kernel should be
    force_kernel_length = int(np.ceil(Fs * N_seconds))
    if force_kernel_length % 2 == 0:
        force_kernel_length = force_kernel_length + 1

    # Check correct length when defining number of seconds
    assert np.allclose(len(kernel), force_kernel_length, atol=.1)



"""
test_spectral.py
Test functions in the spectral domain analysis module
"""

import numpy as np
import os
import neurodsp
from neurodsp import spectral
from neurodsp.tests import _load_example_data


def test_psd():
    """
    Confirm consistency in PSD computation
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    fs = 1000.

    # load "ground truth" PSDs
    gt_psd = np.load(os.path.dirname(neurodsp.__file__) +
                     '/tests/data/sample_data_' + str(data_idx) + '_psd.npz')

    # try all 3 different methods
    freq, Pmean = spectral.psd(x, fs, method='mean', nperseg=fs * 2)
    freq, Pmed = spectral.psd(x, fs, method='median', nperseg=fs * 2)
    freqmf, Pmedfilt = spectral.psd(x, fs, method='medfilt')

    # compute the difference
    assert np.allclose(np.sum(np.abs(gt_psd['PSDmean'] - Pmean)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_psd['PSDmed'] - Pmed)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_psd['PSDmedfilt'] - Pmedfilt)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_psd['freq'] - freq)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_psd['freqmf'] - freqmf)), 0, atol=10 ** -5)


def test_scv():
    """
    Confirm SCV calculation
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    fs = 1000.

    # load ground truth scv
    gt_scv = np.load(os.path.dirname(neurodsp.__file__) +
                     '/tests/data/sample_data_' + str(data_idx) + '_scv.npz')

    # compute SCV
    freq, SCV = spectral.scv(x, fs)
    assert np.allclose(np.sum(np.abs(gt_scv['freq'] - freq)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_scv['SCV'] - SCV)), 0, atol=10 ** -5)


def test_scvrs():
    """
    Confirm SCV resampling calculation
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    fs = 1000.

    # load ground truth scv
    gt_scvrs = np.load(os.path.dirname(neurodsp.__file__) +
                       '/tests/data/sample_data_' + str(data_idx) + '_scvrs.npz')

    # compute SCV and compare differences
    np.random.seed(99)
    freqbs, Tbs, SCVrsbs = spectral.scv_rs(x, fs, method='bootstrap', rs_params=(5, 20))
    assert np.allclose(np.sum(np.abs(gt_scvrs['freqbs'] - freqbs)), 0, atol=10 ** -5)
    assert Tbs is None
    assert np.allclose(np.sum(np.abs(gt_scvrs['SCVrsbs'] - SCVrsbs)), 0, atol=10 ** -5)

    freqro, Tro, SCVrsro = spectral.scv_rs(x, fs, method='rolling', rs_params=(4, 2))
    assert np.allclose(np.sum(np.abs(gt_scvrs['freqro'] - freqro)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_scvrs['Tro'] - Tro)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_scvrs['SCVrsro'] - SCVrsro)), 0, atol=10 ** -5)


def test_spectralhist():
    """
    Confirm SCV resampling calculation
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    fs = 1000.

    # load ground truth scv
    gt_sphist = np.load(os.path.dirname(neurodsp.__file__) +
                        '/tests/data/sample_data_' + str(data_idx) + '_sphist.npz')

    freq, bins, sp_hist = spectral.spectral_hist(x, fs, nbins=10)
    assert np.allclose(np.sum(np.abs(gt_sphist['freq'] - freq)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_sphist['bins'] - bins)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(gt_sphist['sp_hist'] - sp_hist)), 0, atol=10 ** -5)


def test_fitpsd():
    """
    Confirm PSD fitting procedure
    """
    # Load data
    data_idx = 1

    # load test data PSDs for testing
    gt_psd = np.load(os.path.dirname(neurodsp.__file__) +
                     '/tests/data/sample_data_' + str(data_idx) + '_psd.npz')

    psd = gt_psd['PSDmean']
    freq = gt_psd['freq']
    slope_ols, offset_ols = spectral.fit_slope(freq, psd, (30, 100), method='ols')
    slope_rsc, offset_rsc = spectral.fit_slope(freq, psd, (30, 100), method='RANSAC')

    # load ground truth fits
    gt_fitpsd = np.load(os.path.dirname(neurodsp.__file__) +
                        '/tests/data/sample_data_' + str(data_idx) + '_fitpsd.npz')

    assert np.allclose(gt_fitpsd['slope_ols'] - slope_ols, 0, atol=10 ** -5)
    assert np.allclose(gt_fitpsd['slope_rsc'] - slope_rsc, 0, atol=10 ** -5)
    assert np.allclose(gt_fitpsd['offset_ols'] - offset_ols, 0, atol=10 ** -5)
    assert np.allclose(gt_fitpsd['offset_rsc'] - offset_rsc, 0, atol=10 ** -5)



"""
util.py
Utility functions for testing neurodsp functions.
"""

import neurodsp
import os
import numpy as np


def _generate_random_x(len_x=2000, seed=0):
    """Generate a random time series"""
    np.random.seed(seed)
    x = np.random.randn(len_x)
    return x


def _load_example_data(data_idx=1, filtered=False):
    """Load an example voltage time series collected experimentally"""

    # Load time series
    x = np.load(os.path.dirname(neurodsp.__file__) + '/tests/data/sample_data_'+str(data_idx)+'.npy')

    # Load ground-truth filtered data
    if filtered:
        x_filt_true = np.load(os.path.dirname(neurodsp.__file__) + '/tests/data/sample_data_'+str(data_idx)+'_filt.npy')
        return x, x_filt_true
    else:
        return x



"""
phase.py
Estimate the phase of an oscillation using a waveform-based approach
"""

import numpy as np


def extrema_interpolated_phase(x, Ps, Ts, zeroxR=None, zeroxD=None):
    """
    Use peaks (phase 0) and troughs (phase pi/-pi) to estimate
    instantaneous phase. Also use rise and decay zerocrossings
    (phase -pi/2 and pi/2, respectively) if specified.

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Ps : array-like 1d
        time points of oscillatory peaks
    Ts : array-like 1d
        time points of oscillatory troughs
    zeroxR : array-like 1d
        indices at which oscillatory rising zerocrossings occur
    zeroxD : array-like 1d
        indices at which oscillatory decaying zerocrossings occur

    Returns
    -------
    pha : array-like 1d
        instantaneous phase

    Notes
    -----
    Sometimes, due to noise, extrema and zerocrossing estimation
    is poor, and for example, the same index may be assigned to
    both a peak and a decaying zerocrossing. Because of this,
    we first assign phase values by zerocrossings, and then
    may overwrite them with extrema phases.
    """

    # Initialize phase arrays
    # 2 phase arrays: trough pi and trough -pi
    L = len(x)
    t = np.arange(L)
    pha_tpi = np.zeros(L) * np.nan
    pha_tnpi = np.zeros(L) * np.nan

    # If specified, assign phases to zerocrossings
    if zeroxR is not None:
        pha_tpi[zeroxR] = -np.pi / 2
        pha_tnpi[zeroxR] = -np.pi / 2
    if zeroxD is not None:
        pha_tpi[zeroxD] = np.pi / 2
        pha_tnpi[zeroxD] = np.pi / 2

    # Define phases
    pha_tpi[Ps] = 0
    pha_tpi[Ts] = np.pi
    pha_tnpi[Ps] = 0
    pha_tnpi[Ts] = -np.pi

    # Interpolate to find all phases
    pha_tpi = np.interp(t, t[~np.isnan(pha_tpi)], pha_tpi[~np.isnan(pha_tpi)])
    pha_tnpi = np.interp(t, t[~np.isnan(pha_tnpi)], pha_tnpi[~np.isnan(pha_tnpi)])

    # For the phase time series in which the trough is negative pi:
    # Replace the decaying periods with these periods in the phase time
    # series in which the trough is pi
    diffs = np.diff(pha_tnpi)
    diffs = np.append(diffs, 99)
    pha_tnpi[diffs < 0] = pha_tpi[diffs < 0]

    # Assign the periods before the first empirical phase timepoint to NaN
    diffs = np.diff(pha_tnpi)
    first_empirical_idx = next(i for i, xi in enumerate(diffs) if xi > 0)
    pha_tnpi[:first_empirical_idx] = np.nan

    # Assign the periods after the last empirical phase timepoint to NaN
    diffs = np.diff(pha_tnpi)
    last_empirical_idx = next(i for i, xi in enumerate(diffs[::-1]) if xi > 0)
    pha_tnpi[-last_empirical_idx + 1:] = np.nan

    return pha_tnpi



from .filt import filter
from .pac import compute_pac, compute_pac_comodulogram, plot_pac_comodulogram
from .laggedcoherence import lagged_coherence
from .timefrequency import phase_by_time, amp_by_time, freq_by_time
from . import shape
from . import spectral



"""
test_swm.py
Test the sliding window matching function
"""

import numpy as np
import os
import neurodsp
from neurodsp import shape
from neurodsp.tests import _load_example_data


def test_swm_consistent():
    """
    Confirm consistency in beta bandpass filter results on a neural signal
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000

    # Load ground truth lagged coherence
    avg_window_true = np.load(os.path.dirname(neurodsp.__file__) +
                              '/tests/data/sample_data_' + str(data_idx) + '_swm.npy')

    # Compute lagged coherence
    L = .055
    G = .2
    np.random.seed(1)
    avg_window, _, _ = shape.sliding_window_matching(x, Fs, L, G, max_iterations=500)

    # Compute difference between current and past signals
    signal_diff = avg_window - avg_window_true
    assert np.allclose(np.sum(np.abs(signal_diff)), 0, atol=10 ** -5)



"""
test_shapephase.py
Test the phase estimation technique based on extrema and zerocrossings
"""

import numpy as np
import os
import neurodsp
from neurodsp import shape
from neurodsp.tests import _load_example_data


def test_interpolated_phase_consistent():
    """
    Confirm consistency in beta bandpass filter results on a neural signal
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000
    f_range = (13, 30)

    # Load ground truth phase time series
    phaPTRD_true = np.load(os.path.dirname(neurodsp.__file__) +
                           '/tests/data/sample_data_' + str(data_idx) + '_phaPTRD.npy')

    # Compute phase time series
    Ps, Ts = shape.find_extrema(x, Fs, f_range)
    zeroxR, zeroxD = shape.find_zerox(x, Ps, Ts)
    phaPTRD = shape.extrema_interpolated_phase(x, Ps, Ts, zeroxR=zeroxR, zeroxD=zeroxD)

    # Compute difference between current and past filtered signals
    signal_diff = phaPTRD[~np.isnan(phaPTRD)] - phaPTRD_true[~np.isnan(phaPTRD_true)]
    assert np.allclose(np.sum(np.abs(signal_diff)), 0, atol=10 ** -5)



"""
laggedcoherence.py
Estimate the rhythmicity of a neural oscillation by calculating the lagged coherence metric
"""

import numpy as np
from scipy import signal


def lagged_coherence(x, f_range, Fs, N_cycles=3, f_step=1, return_spectrum=False):
    """
    Quantify the rhythmicity of an oscillator using the lagged coherence measure.

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    f_range : (low, high), Hz
        frequency range of the oscillator
    Fs : float
        sampling rate
    N_cycles : float
        Number of cycles of the frequency of interest used to compute lagged coherence
    f_step : float, Hz
        step size to calculate lagged coherence in the frequency range.
    return_spectrum : bool
        if True, return the lagged coherence for all frequency values. otherwise, only return mean

    Returns
    -------
    lc : float (or numpy.array 1d)
        if return_spectrum is False: mean lagged coherence value in the frequency range of interest
        if return_spectrum is True: lagged coherence value for each frequency in the frequency range
    freqs : numpy.array 1d
        Only returned if return_spectrum is True
        Frequencies (Hz) corresponding to the lagged coherence values in lc

    References
    ----------
    Fransen, A. M., van Ede, F., & Maris, E. (2015).
    Identifying neuronal oscillations using rhythmicity.
    Neuroimage, 118, 256-267.
    """

    # Identify Fourier components of interest
    freqs = np.arange(f_range[0], f_range[1] + f_step, f_step)

    # Calculate lagged coherence for each frequency
    F = len(freqs)
    lcs = np.zeros(F)
    for i, f in enumerate(freqs):
        lcs[i] = _lagged_coherence_1freq(
            x, f, Fs, N_cycles=N_cycles, f_step=f_step)

    # Return desired measure of lagged coherence
    if return_spectrum:
        lc = lcs
        return lc, freqs
    else:
        lc = np.mean(lcs)
        return lc


def _lagged_coherence_1freq(x, f, Fs, N_cycles=3, f_step=1):
    """Calculate lagged coherence of x at frequency f using the hanning-taper FFT method"""

    # Determine number of samples to be used in each window to compute lagged coherence
    Nsamp = int(np.ceil(N_cycles * Fs / f))

    # For each N-cycle chunk, calculate the fourier coefficient at the frequency of interest, f
    chunks = _nonoverlapping_chunks(x, Nsamp)
    C = len(chunks)
    hann_window = signal.hanning(Nsamp)
    fourier_f = np.fft.fftfreq(Nsamp, 1 / float(Fs))
    fourier_f_idx = np.argmin(np.abs(fourier_f - f))
    fourier_coefsoi = np.zeros(C, dtype=complex)
    for i2, c in enumerate(chunks):
        fourier_coef = np.fft.fft(c * hann_window)
        fourier_coefsoi[i2] = fourier_coef[fourier_f_idx]

    # Compute the lagged coherence value
    lcs_num = 0
    for i2 in range(C - 1):
        lcs_num += fourier_coefsoi[i2] * np.conj(fourier_coefsoi[i2 + 1])
    lcs_denom = np.sqrt(np.sum(
        np.abs(fourier_coefsoi[:-1])**2) * np.sum(np.abs(fourier_coefsoi[1:])**2))
    return np.abs(lcs_num / lcs_denom)


def _nonoverlapping_chunks(x, N):
    """Split x into nonoverlapping chunks of length N"""
    Nchunks = int(np.floor(len(x) / float(N)))
    chunks = np.reshape(x[:int(Nchunks * N)], (Nchunks, int(N)))
    return chunks



"""
filt.py
Filter a neural signal using a bandpass, highpass, lowpass, or bandstop filter.
"""

import numpy as np
import scipy as sp
from scipy import signal
import matplotlib.pyplot as plt
import warnings


def filter(x, Fs, pass_type, f_lo=None, f_hi=None, N_cycles=3, N_seconds=None,
           iir=False, butterworth_order=None,
           plot_frequency_response=False, return_kernel=False,
           verbose=True, compute_transition_band=True, remove_edge_artifacts=True):
    """
    Apply a bandpass, bandstop, highpass, or lowpass filter to a neural signal

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Fs : float
        The sampling rate
    pass_type : str
        'bandpass' : apply a bandpass filter
        'bandstop' : apply a bandstop (notch) filter
        'lowpass' : apply a lowpass filter
        'highpass' : apply a highpass filter
    f_lo : float
        Low-frequency cutoff (Hz)
    f_hi : float
        High-frequency cutoff (Hz)
    N_cycles : float
        Length of filter in terms of number of cycles at 'f_lo' frequency
        This parameter is overwritten by 'N_seconds'
    N_seconds : float
        Length of filter (seconds)
    iir : bool
        if True, use an infinite-impulse response (IIR) filter
        The only IIR filter to be used is a butterworth filter
    butterworth_order : int
        order of the butterworth filter
        see input 'N' in scipy.signal.butter
    plot_frequency_response : bool
        if True, plot the frequency response of the filter
    return_kernel : bool
        if True, return the complex filter kernel
    verbose : bool
        if True, print optional information
    compute_transition_band : bool
        if True, the function computes the transition bandwidth,
        defined as the frequency range between -20dB and -3dB attenuation,
        and warns the user if this band is longer than the frequency bandwidth.
    remove_edge_artifacts : bool
        if True, replace the samples that are within half a kernel's length to
        the signal edge with np.nan

    Returns
    -------
    x_filt : array-like 1d
        filtered time series
    kernel : length-2 tuple of arrays
        filter kernel
        returned only if 'return_kernel' == True
    """

    # Check that frequency cutoff inputs are appropriate
    if pass_type == 'bandpass' or pass_type == 'bandstop':
        if f_lo is None or f_hi is None:
            raise ValueError('For bandpass and bandstop filters, both f_lo and f_hi must be defined.')
        if f_lo >= f_hi:
            raise ValueError('f_lo must be less than f_hi.')

    elif pass_type == 'highpass':
        if f_lo is not None:
            raise ValueError('For highpass filters, f_lo must not be used. f_hi is used to indicate the cutoff frequency.')
        if f_hi is None:
            raise ValueError('For highpass filters, f_hi must be used to represent the cutoff frequency. f_lo is not used')

    elif pass_type == 'lowpass':
        if f_hi is not None:
            raise ValueError('For lowpass filters, f_hi must not be used. f_lo is used to indicate the cutoff frequency.')
        if f_lo is None:
            raise ValueError('For lowpass filters, f_lo must be used to represent the cutoff frequency. f_hi is not used')

    else:
        raise ValueError('Input for "pass_type" not recognized. Must indicate bandpass, bandstop, lowpass, or highpass.')

    # Remove any NaN on the edges of 'x'
    first_nonan = np.where(~np.isnan(x))[0][0]
    last_nonan = np.where(~np.isnan(x))[0][-1] + 1
    x_old = np.copy(x)
    x = x[first_nonan:last_nonan]

    # Process input for IIR filters
    if iir:
        if remove_edge_artifacts:
            if verbose:
                warnings.warn('Edge artifacts are not removed when using an IIR filter.')
        if pass_type != 'bandstop':
            if verbose:
                warnings.warn('IIR filters are not recommended other than for notch filters.')
        if N_seconds is not None:
            raise TypeError('N_seconds should not be defined for an IIR filter.')
        if butterworth_order is None:
            raise TypeError('butterworth_order must be defined when using an IIR filter.')

    # Process input for FIR filters
    else:
        # Compute filter length if specified in seconds
        if N_seconds is not None:
            N = int(np.ceil(Fs * N_seconds))
        else:
            if pass_type == 'highpass':
                N = int(np.ceil(Fs * N_cycles / f_hi))
            else:
                N = int(np.ceil(Fs * N_cycles / f_lo))

        # Force filter length to be odd
        if N % 2 == 0:
            N = int(N + 1)

    # Compute nyquist frequency
    f_nyq = Fs / 2.

    # Design filter
    if iir:
        if pass_type == 'bandpass' or pass_type == 'bandstop':
            Wn = (f_lo / f_nyq, f_hi / f_nyq)
        elif pass_type == 'highpass':
            Wn = f_hi / f_nyq
        elif pass_type == 'lowpass':
            Wn = f_lo / f_nyq
        b, a = sp.signal.butter(butterworth_order, Wn, pass_type)
    else:
        if pass_type == 'bandpass':
            kernel = sp.signal.firwin(N, (f_lo, f_hi), pass_zero=False, nyq=f_nyq)
        elif pass_type == 'bandstop':
            kernel = sp.signal.firwin(N, (f_lo, f_hi), nyq=f_nyq)
        elif pass_type == 'highpass':
            kernel = sp.signal.firwin(N, f_hi, pass_zero=False, nyq=f_nyq)
        elif pass_type == 'lowpass':
            kernel = sp.signal.firwin(N, f_lo, nyq=f_nyq)

    # Apply filter
    if iir:
        x_filt = sp.signal.filtfilt(b, a, x)
    else:
        x_filt = np.convolve(kernel, x, 'same')

    # Plot frequency response, if desired
    if plot_frequency_response:
        if iir:
            _plot_frequency_response(Fs, b, a)
        else:
            _plot_frequency_response(Fs, kernel)

    # Compute transition bandwidth
    if compute_transition_band and verbose:

        # Compute the frequency response in terms of Hz and dB
        if not iir:
            b = kernel
            a = 1
        w, h = signal.freqz(b, a)
        f_db = w * Fs / (2. * np.pi)
        db = 20 * np.log10(abs(h))

        # Compute pass bandwidth and transition bandwidth
        if pass_type == 'bandpass':
            pass_bw = f_hi - f_lo
            # Identify edges of transition band (-3dB and -20dB)
            cf_20db_1 = next(f_db[i] for i in range(len(db)) if db[i] > -20)
            cf_3db_1 = next(f_db[i] for i in range(len(db)) if db[i] > -3)
            cf_20db_2 = next(f_db[i] for i in range(len(db))[::-1] if db[i] > -20)
            cf_3db_2 = next(f_db[i] for i in range(len(db))[::-1] if db[i] > -3)
            # Compute transition bandwidth
            transition_bw1 = cf_3db_1 - cf_20db_1
            transition_bw2 = cf_20db_2 - cf_3db_2
            transition_bw = max(transition_bw1, transition_bw2)

            if cf_20db_1 == f_db[0]:
                warnings.warn('The low frequency stopband never gets attenuated by more than 20dB. Increase filter length.')
            if cf_20db_2 == f_db[-1]:
                warnings.warn('The high frequency stopband never gets attenuated by more than 20dB. Increase filter length.')

        elif pass_type == 'bandstop':
            pass_bw = f_hi - f_lo
            # Identify edges of transition band (-3dB and -20dB)
            cf_20db_1 = next(f_db[i] for i in range(len(db)) if db[i] < -20)
            cf_3db_1 = next(f_db[i] for i in range(len(db)) if db[i] < -3)
            cf_20db_2 = next(f_db[i] for i in range(len(db))[::-1] if db[i] < -20)
            cf_3db_2 = next(f_db[i] for i in range(len(db))[::-1] if db[i] < -3)
            # Compute transition bandwidth
            transition_bw1 = cf_20db_1 - cf_3db_1
            transition_bw2 = cf_3db_2 - cf_20db_2
            transition_bw = max(transition_bw1, transition_bw2)

        elif pass_type == 'highpass':
            pass_bw = f_hi
            # Identify edges of transition band (-3dB and -20dB)
            cf_20db = next(f_db[i] for i in range(len(db)) if db[i] > -20)
            cf_3db = next(f_db[i] for i in range(len(db)) if db[i] > -3)
            # Compute transition bandwidth
            transition_bw = cf_3db - cf_20db

        elif pass_type == 'lowpass':
            pass_bw = f_lo
            # Identify edges of transition band (-3dB and -20dB)
            cf_20db = next(f_db[i] for i in range(len(db)) if db[i] < -20)
            cf_3db = next(f_db[i] for i in range(len(db)) if db[i] < -3)
            # Compute transition bandwidth
            transition_bw = cf_20db - cf_3db

        # Raise warning if transition bandwidth is too high
        if transition_bw > pass_bw:
            warnings.warn('Transition bandwidth is ' + str(np.round(transition_bw, 1)) + ' Hz. This is greater than the desired pass/stop bandwidth of ' + str(np.round(pass_bw, 1)) + ' Hz')

    # Remove edge artifacts
    if not iir and remove_edge_artifacts:
        N_rmv = int(np.ceil(N / 2))
        x_filt[:N_rmv] = np.nan
        x_filt[-N_rmv:] = np.nan

    # Add NaN back on the edges of 'x', if there were any at the beginning
    x_filt_full = np.ones(len(x_old)) * np.nan
    x_filt_full[first_nonan:last_nonan] = x_filt
    x_filt = x_filt_full

    # Return kernel if desired
    if return_kernel:
        if iir:
            return x_filt, (b, a)
        else:
            return x_filt, kernel
    else:
        return x_filt


def _plot_frequency_response(Fs, b, a=1):
    """Compute frequency response of a filter kernel b with sampling rate Fs"""
    w, h = signal.freqz(b, a)
    # Plot frequency response
    plt.figure(figsize=(10, 5))
    plt.subplot(1, 2, 1)
    plt.plot(w * Fs / (2. * np.pi), 20 * np.log10(abs(h)), 'k')
    plt.title('Frequency response')
    plt.ylabel('Attenuation (dB)')
    plt.xlabel('Frequency (Hz)')
    if type(a) is int:
        # Plot filter kernel
        plt.subplot(1, 2, 2)
        plt.plot(b, 'k')
        plt.title('Kernel')
    plt.show()



"""
test_cyclepoints.py
Test identification of extrema and zerocrossings
"""

import numpy as np
import os
import neurodsp
from neurodsp import shape
from neurodsp.tests import _load_example_data


def test_Ps_consistent():
    """
    Confirm consistency in peak finding
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000
    f_range = (13, 30)

    # Load ground truth lagged coherence
    Ps_true = np.load(os.path.dirname(neurodsp.__file__) +
                      '/tests/data/sample_data_' + str(data_idx) + '_Ps.npy')

    # Compute lagged coherence
    Ps, Ts = shape.find_extrema(x, Fs, f_range)

    # Compute difference between current and past signals
    signal_diff = Ps - Ps_true
    assert np.allclose(np.sum(np.abs(signal_diff)), 0, atol=10 ** -5)


def test_zeroxR_consistent():
    """
    Confirm consistency in peak finding
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000
    f_range = (13, 30)

    # Load ground truth lagged coherence
    zeroxR_true = np.load(os.path.dirname(neurodsp.__file__) +
                          '/tests/data/sample_data_' + str(data_idx) + '_zeroxR.npy')

    # Compute lagged coherence
    Ps, Ts = shape.find_extrema(x, Fs, f_range)
    zeroxR, zeroxD = shape.find_zerox(x, Ps, Ts)

    # Compute difference between current and past signals
    signal_diff = zeroxR - zeroxR_true
    assert np.allclose(np.sum(np.abs(signal_diff)), 0, atol=10 ** -5)




"""neurodsp setup script"""
from setuptools import setup, find_packages
setup(
    name = 'neurodsp',
    version = '0.2',
    description = 'A package of modules for analyzing neural signals',
    author = 'The Voytek Lab',
    author_email = 'voyteklab@gmail.com',
    url = 'https://github.com/voytekresearch/neurodsp',
    packages=find_packages(),
    license='MIT',
    download_url = 'https://github.com/voytekresearch/neurodsp/archive/0.2.tar.gz',
    keywords = ['neuroscience', 'neural oscillations', 'time series analysis', 'spectral analysis', 'LFP'],
    classifiers = []
)



"""
timefrequency.py
Tools for estimating properties of a neural oscillation over time
"""

import numpy as np
import scipy as sp
from scipy import signal
import math
import neurodsp


def phase_by_time(x, Fs, f_range,
                  filter_fn=None, filter_kwargs=None,
                  hilbert_increase_N=False):
    """
    Calculate the phase time series of a neural oscillation

    Parameters
    ----------
    x : array-like, 1d
        Time series
    f_range : (low, high), Hz
        Frequency range
    Fs : float, Hz
        Sampling rate
    filter_fn : function
        The filtering function, with api:
        `filterfn(x, Fs, pass_type, f_lo, f_hi, remove_edge_artifacts=True)
    filter_kwargs : dict
        Keyword parameters to pass to `filterfn(.)`
    hilbert_increase_N : bool
        if True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of x

    Returns
    -------
    pha : array-like, 1d
        Time series of phase
    """
    # Set default filtering parameters
    if filter_fn is None:
        filter_fn = neurodsp.filter
    if filter_kwargs is None:
        filter_kwargs = {}
    # Filter signal
    x_filt = filter_fn(x, Fs, 'bandpass', f_lo=f_range[0], f_hi=f_range[1], remove_edge_artifacts=False, **filter_kwargs)
    # Compute phase time series
    pha = np.angle(_hilbert_ignore_nan(x_filt, hilbert_increase_N=hilbert_increase_N))
    return pha


def amp_by_time(x, Fs, f_range,
                filter_fn=None, filter_kwargs=None,
                hilbert_increase_N=False):
    """
    Calculate the amplitude time series

    Parameters
    ----------
    x : array-like, 1d
        Time series
    f_range : (low, high), Hz
        The frequency filtering range
    Fs : float, Hz
        Sampling rate
    filter_fn : function
        The filtering function, `filterfn(x, f_range, filter_kwargs)`
        Must have the same API as filt.bandpass
    filter_kwargs : dict
        Keyword parameters to pass to `filterfn(.)`
    hilbert_increase_N : bool
        if True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of x

    Returns
    -------
    amp : array-like, 1d
        Time series of amplitude
    """
    # Set default filtering parameters
    if filter_fn is None:
        filter_fn = neurodsp.filter
    if filter_kwargs is None:
        filter_kwargs = {}
    # Filter signal
    x_filt = filter_fn(x, Fs, 'bandpass', f_lo=f_range[0], f_hi=f_range[1], remove_edge_artifacts=False, **filter_kwargs)
    # Compute amplitude time series
    amp = np.abs(_hilbert_ignore_nan(x_filt, hilbert_increase_N=hilbert_increase_N))
    return amp


def freq_by_time(x, Fs, f_range):
    '''
    Estimate the instantaneous frequency at each sample

    Parameters
    ----------
    x : array-like 1d
        voltage time series
    Fs : float
        sampling rate
    f_range : (low, high), Hz
        frequency range for filtering

    Returns
    -------
    i_f : float
        estimate instantaneous frequency for each sample in 'x'

    Notes
    -----
    * This function assumes monotonic phase, so
    a phase slip will be processed as a very high frequency
    '''
    pha = phase_by_time(x, Fs, f_range)
    phadiff = np.diff(pha)
    phadiff[phadiff < 0] = phadiff[phadiff < 0] + 2 * np.pi
    i_f = Fs * phadiff / (2 * np.pi)
    i_f = np.insert(i_f, 0, np.nan)
    return i_f


def _hilbert_ignore_nan(x, hilbert_increase_N=False):
    """
    Compute the hilbert transform of x.
    Ignoring the boundaries of x that are filled with NaN
    """
    # Extract the signal that is not nan
    first_nonan = np.where(~np.isnan(x))[0][0]
    last_nonan = np.where(~np.isnan(x))[0][-1] + 1
    x_nonan = x[first_nonan:last_nonan]

    # Compute hilbert transform of signal without nans
    if hilbert_increase_N:
        N = len(x_nonan)
        N2 = 2**(int(math.log(N, 2)) + 1)
        x_hilb_nonan = signal.hilbert(x_nonan, N2)
    else:
        x_hilb_nonan = signal.hilbert(x_nonan)

    # Fill in output hilbert with nans on edges
    x_hilb = np.ones(len(x), dtype=complex) * np.nan
    x_hilb[first_nonan:last_nonan] = x_hilb_nonan
    return x_hilb



"""
test_timefreq.py
Test functions in the time-frequency analysis module
"""

import numpy as np
import os
import neurodsp
from neurodsp.tests import _load_example_data


def test_timefreq_consistent():
    """
    Confirm consistency in estimation of instantaneous phase, amp, and frequency
    with computations in previous versions
    """
    # Load data
    data_idx = 1
    x = _load_example_data(data_idx=data_idx)
    Fs = 1000
    f_range = (13, 30)

    # Load ground truth phase time series
    pha_true = np.load(os.path.dirname(neurodsp.__file__) +
                       '/tests/data/sample_data_' + str(data_idx) + '_pha.npy')
    # Load ground truth amplitude time series
    amp_true = np.load(os.path.dirname(neurodsp.__file__) +
                       '/tests/data/sample_data_' + str(data_idx) + '_amp.npy')
    # Load ground truth frequency time series
    i_f_true = np.load(os.path.dirname(neurodsp.__file__) +
                       '/tests/data/sample_data_' + str(data_idx) + '_i_f.npy')

    # Compute phase time series
    pha = neurodsp.phase_by_time(x, Fs, f_range)
    # Compute amplitude time series
    amp = neurodsp.amp_by_time(x, Fs, f_range)
    # Compute frequency time series
    i_f = neurodsp.freq_by_time(x, Fs, f_range)

    # Compute difference between current and past signals
    assert np.allclose(np.sum(np.abs(pha - pha_true)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(amp - amp_true)), 0, atol=10 ** -5)
    assert np.allclose(np.sum(np.abs(i_f[~np.isnan(i_f)] - i_f_true[~np.isnan(i_f_true)])),
                       0, atol=10 ** -5)


def test_NaN_in_x():
    """
    Assure that time-resolved timefrequency functions do not return all NaN
    if one of the elements in the input array is NaN.
    Do this by replacing edge artifacts with NaN for a lowpass filter
    """

    # Generate a low-pass filtered signal with NaNs
    x = np.random.randn(10000)
    Fs = 1000
    x = neurodsp.filter(x, Fs, 'lowpass', f_lo=50)

    # Compute phase, amp, and freq time series
    f_range = (4, 8)
    pha = neurodsp.phase_by_time(x, Fs, f_range)
    amp = neurodsp.amp_by_time(x, Fs, f_range)
    i_f = neurodsp.freq_by_time(x, Fs, f_range)

    assert len(pha[~np.isnan(pha)]) > 0
    assert len(amp[~np.isnan(amp)]) > 0
    assert len(i_f[~np.isnan(i_f)]) > 0
